#include <stdio.h>
#include <string.h>
#include <fstream.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <strclass.h>
#include <ctype.h>

#include "gdbm.h"

/******************************************************************
**  How I compile this program:                                  **
**  1) Rename "cgic.c" to "cgic.cpp" so it will link properly    **
**  2) g++ forum.cpp cgic.cpp libgdbm.a -oforum.cgi              **
**                                                               **
**  On my system it compiles cleanly if done this way.  If you   **
**  get link errors, it's because you compiled cgic.c as a C     **
**  program and forum.cpp as a C++ program.                      **
**                                                               **
**  I have compiled this successfully with DJGPP under           **
**  DOS/Windows 95 and with g++ under BSDI Unix.  I am unable to **
**  offer support for free software such as this.  It was        **
**  written for my own use, and I am making it available in the  **
**  hope that it may be useful to others, but with no waranty at **
**  all.  I will not take responsibility for any damage this     **
**  program may cause to your system.                            **
**								 **
**  See comments below for information about CGIC.CPP.           **
**                                                               **
**  Please see the file readme.txt which is available at         **
**  http://www.writersbbs.com/cgi-src/forum/readme.txt for       **
**  important information you will need to install this program. **
*******************************************************************
*/

// RCS Log follows
//$Log: forum.cpp $
//Revision 1.16  1997/03/29 17:43:38  brian
//Added comments for installation help.
//
//Revision 1.15  1997/03/29 16:38:13  brian
//added RCS log info
//

#include "cgic.h"
/*
cgic is obtainable from http://www.boutell.com/cgic/

cgic, copyright 1996 by Thomas Boutell. Permission is granted
to use cgic in any application, commercial or noncommercial,
at no cost. HOWEVER, this copyright paragraph must appear on
a "credits" page accessible in the public online and offline
documentation of the program. Modified versions of the cgic library
should not be distributed without the attachment of a clear
statement regarding the author of the modifications, and this
notice may in no case be removed. Modifications may also be
submitted to the author for inclusion in the main cgic distribution. 
*/

// Following are settings you will need to change for your site
const char SALT[]="gd"; // used with crypt
const char *EntraceURL="/your_bbs/"; // URL of the entrance/index page
const char *cgiURL="/cgi-bin/your_bbs_script_directory/forum.cgi";	// URL of this program
const char *WebmasterEmail="webmaster@yoursite.com";	// Your e-mail address
char *PasswordFileName="/your_private_directory/pw.gdb"; // Full path to the password database file

// Following should not need to be changed

// contents of this file are displayed at the top of most screens
const char *commonHeaderFileName="common.h"; 

const int MaxFileNameLength=50;		// Max size of a file name
const int MaxUserNameLength=30;		// Max size of a user name
const int MaxSubjectLength=80;
struct UserInfo
{
	char pass[MaxUserNameLength];
	time_t lastLogin;
};

struct MessageNumber
{
	unsigned message;
	unsigned reply;
};

struct MessageHeader
{
	char userName[MaxUserNameLength];
	time_t timePosted;
	char subject[MaxSubjectLength];
	MessageNumber number;
};

extern "C" char *crypt(const char *, const char *);

// If your compiler's standard library doesn't have a crypt() function,
// you can add your own here or simply return pw.
char *wrapCrypt(const char *pw)
{
	return crypt(pw, SALT);
}

// If your compiler's standard library has a strupr() function,
// you may want to replace the guts of this function with a call
// to it.
char *WBBS_strupr(char *in)
{
	if (in)
	{
		while (*in)
		{
			*in=toupper(*in);
			in++;
		}
	}
	return in;
}

int compareMessageHeader(const void *h1, const void *h2)
{
	const MessageHeader *mh1 = (const MessageHeader *) h1;
	const MessageHeader *mh2 = (const MessageHeader *) h2;

	int m = mh1->number.message - mh2->number.message;
	return m ? m : mh1->number.reply - mh2->number.reply;	
}

class Lock
{
protected:
	char m_fileName[MaxFileNameLength];
	ifstream m_file;
	int m_locked;
public:
	Lock(const char *fileName, int maxWait);
	virtual ~Lock();
	int locked() {return m_locked;}
	void Unlock();
};

Lock::Lock(const char *name, int maxWait)
{
	strncpy(m_fileName, name, MaxFileNameLength-1);
	m_fileName[MaxFileNameLength-1]='\0';
	int success=0, done=0;
	time_t end = time(NULL) + maxWait; // wait maxWait seconds
	do
	{
		m_file.open(name,ios::noreplace);
		if (!m_file)
		{
			// can't open the lock file
			time_t now=time(NULL);
			if (now > end)
			{
				// Something's wrong, so delete the file and reset the timer
				unlink(m_fileName);
				end = time(NULL) + maxWait;
			}
		}
		else
		{
			// opened the lock file
			m_locked=1;
		}
	} while (!m_locked);
}

void Lock::Unlock()
{
	if (m_file)
	{
		m_file.close();
	}
	unlink(m_fileName);
	m_locked=0;
}

Lock::~Lock()
{
	Unlock();
}

void ShowFile(const char *fileName)
{
	FILE *textFile=fopen(fileName, "r");
	if (textFile)
	{
		char buff[100];
		while (!feof(textFile))
		{
			size_t bytesRead = fread(buff, sizeof(char), 100, textFile);
			fwrite(buff, sizeof(char), bytesRead, cgiOut);
		}
		fclose(textFile);
	}
}

void DoHeader(const char *title)
{
    	fprintf(cgiOut, "Content-type: text/html\n\n");      
	fprintf(cgiOut, "<HTML>");
	fprintf(cgiOut, "<HEAD>\n");
      	fprintf(cgiOut, "<TITLE>%s</TITLE></HEAD>\n", title);
      	fprintf(cgiOut, "<BODY bgcolor=\"#ffffff\">\n");
}

void DoError(const char *message)
{
	const char *title="An error occurred";
	DoHeader(title);
	ShowFile(commonHeaderFileName);
	fprintf(cgiOut, "<H1 align=center>%s</H1>\n", title);
	fprintf(cgiOut, "Please report the following error to the webmaster (<A HREF=\"mailto:%s\">%s</A>)", WebmasterEmail, WebmasterEmail);
	fprintf(cgiOut, "<P><B>%s</B>",message);
	fprintf(cgiOut, "</body></html>");
	exit(0);
}	

void WrongPassword()
{
	DoHeader("Please login");
	ShowFile(commonHeaderFileName);
	fprintf(cgiOut, "<H1>Please login</H1>");
	fprintf(cgiOut, "Your username and password do not match.<P>");
	fprintf(cgiOut, "Please create an account if "
			"you don't have one or have forgotten your password.<P>");
}

int VerifyRoomName(const char *roomName)
{
	FILE *roomFile = fopen("forums.txt", "rt");
	if (roomFile)
	{
		while (!feof(roomFile))
		{
			char buff[MaxFileNameLength];
			fgets(buff, MaxFileNameLength, roomFile);
			char *temp = strtok(buff, "\r\n\t \0");
			if (temp && *temp != '/') // not a blank line or comment
			{
				if (!strcmp(temp,roomName))
				{
					return 1;
				}
			}
		}
	}
	else
	{
		DoError("Could not open configuration file for reading.");
	}
	
	return 0; // not valid
}

String stripHTML(String text, int length)
{
	text.gsub("<", "&lt;");	
	text.gsub(">", "&gt;");
	text.gsub("\"", "&quot;");
	return text.length() > (length-1) ? (String) text.at(0, length-1) : text;
}

void stripHtmlInPlace(char *text)
{
	char *p=text;
	while(*p)
	{
		if (*p == '<')
		{
			*p = '(';
		}
		else if (*p == '>')
		{
			*p = ')';
		}
		else if (*p == '"')
		{
			*p = '\'';
		}
		p++;
	}
}

int VerifyPassword(char *user_name, char *password, int login)
{
	int return_value = 0;

	int flags=GDBM_READER;
	if (login)
	{
		// Need to update the record with the current time
		flags=GDBM_WRITER;
	}
		
	GDBM_FILE pw_file = gdbm_open(PasswordFileName, 0, flags, 00664, NULL);					
	if (!pw_file)
	{
		DoError("Error: could not open password database.");
		return 0;
	}
			
	datum key, data;	
	key.dptr=user_name;
	key.dsize=strlen(key.dptr) + 1;
	data = gdbm_fetch(pw_file, key);
	if (data.dptr)
	{
		UserInfo *user = (UserInfo*) data.dptr;
		if (!strcmp(wrapCrypt(password), user->pass))
		{
			// passwords match
			return_value=1;
			if (login)
			{
				user->lastLogin = time(NULL);
				if (gdbm_store (pw_file, key, data, GDBM_REPLACE))
				{
					// could not insert entry
					DoError("Could not write to the login database");
					return_value=0;
				}
			}	
		}
		free(data.dptr);		
	}
		
	gdbm_close(pw_file);
	return return_value;
}

void CreateAccount()
{
	DoHeader("Password Administration");
	fprintf(cgiOut,"<H1 align=center>Password Administration - Create a new account</H1>\n");

    char name[MaxUserNameLength];
    char pass1[MaxUserNameLength];
    char pass2[MaxUserNameLength];
	*name=*pass1=*pass2='\0';
	
    cgiFormStringNoNewlines("name", name, MaxUserNameLength);
    cgiFormStringNoNewlines("pass1", pass1, MaxUserNameLength);
    cgiFormStringNoNewlines("pass2", pass2, MaxUserNameLength);

	if (!*name || !*pass1 || !*pass2)
	{
		fprintf(cgiOut,	"You did not fill in all the fields."
				" Please use the back button on your web browser to return to the form.");
	}
	else if (strcmp(pass1, pass2))
	{
			fprintf(cgiOut,	"You did not enter the same password twice."
					" Please use the back button on your web browser to return to the form.");
	}
	else
	{
		// Create the account
		GDBM_FILE pw_file = gdbm_open(PasswordFileName, 0, GDBM_WRCREAT, 00664, NULL);					
		if (!pw_file)
		{
			fprintf(cgiOut,	"Error: could not open database.\n");
			return;
		}
			
		datum key, data;	

		UserInfo user;
		strcpy(user.pass, wrapCrypt(pass1));
		user.lastLogin = time(NULL);
		data.dptr= (char *) &user;
		data.dsize=sizeof (UserInfo);

		stripHtmlInPlace(name);
		key.dptr=name;
		key.dsize=strlen(name) + 1;
		
		if (gdbm_store (pw_file, key, data, GDBM_INSERT))
		{
			// could not insert entry
			fprintf(cgiOut,	"Sorry, that name is already in use."
					" Please use the back button on your web browser to return to the form.");
		}
		else
		{
			fprintf(cgiOut,	"Your account has been created.  You may now login to the forums"
					" with your username (<strong>%s</strong>) and password (<strong>%s</strong>).  Please write them down."
					" If you forget your password, you will have to create a new account.", name, pass1);
		}
		gdbm_close(pw_file);
	}
}

void ChangePassword()
{
	DoHeader("Password Administration");
	fprintf(cgiOut,"<H1 align=center>Password Administration - Change your password</H1>\n");

	char name[MaxUserNameLength];
	char OldPass[MaxUserNameLength];
	char NewPass1[MaxUserNameLength];
	char NewPass2[MaxUserNameLength];
	*name=*OldPass=*NewPass1=*NewPass2='\0';

	datum key, data;	
	
	cgiFormStringNoNewlines("name", name, MaxUserNameLength);
	cgiFormStringNoNewlines("old_pass", OldPass, MaxUserNameLength);
	cgiFormStringNoNewlines("new_pass1", NewPass1, MaxUserNameLength);
	cgiFormStringNoNewlines("new_pass2", NewPass2, MaxUserNameLength);

	if (!*name || !*OldPass || !*NewPass1 || !*NewPass2)
	{
		fprintf(cgiOut,	"You did not fill in all the fields."
				" Please use the back button on your web browser to return to the form.");
	}
	else if (strcmp(NewPass1, NewPass2))
	{
		fprintf(cgiOut,	"You did not enter the same New Password password twice."
				" Please use the back button on your web browser to return to the form.");
	}
	else
	{
		GDBM_FILE pw_file = gdbm_open(PasswordFileName, 0, GDBM_WRCREAT, 00664, NULL);					
		if (!pw_file)
		{
			fprintf(cgiOut,	"Error: could not open database.\n");
			return;
		}

		// verify that they entered the correct password
		key.dptr=name;
		key.dsize=strlen(key.dptr) + 1;
		data = gdbm_fetch(pw_file, key);
		if (!data.dptr)
		{
			fprintf(cgiOut, "That account doesn't exist."
					" Please use the back button on your web browser to return to the form.");
		}
		else
		{
			UserInfo *user=NULL;
			user = (UserInfo *) data.dptr;		
			if (strcmp(user->pass, wrapCrypt(OldPass)))
			{
				fprintf(cgiOut, "Wrong password."
						" Please use the back button on your web browser to return to the form.");
			}
			else
			{
				strcpy(user->pass, wrapCrypt(NewPass1));
				user->lastLogin = time(NULL);

				if (gdbm_store (pw_file, key, data, GDBM_REPLACE))
				{
					// could not insert entry
					fprintf(cgiOut,	"Error: could not write to the database.\n");				
				}
				else
				{
					fprintf(cgiOut,	"Your password has been changed.  You may now login to the forums"
					" with your username (<strong>%s</strong>) and password (<strong>%s</strong>).  Please write them down."
					" If you forget your password, you will have to create a new account.", name, NewPass1);
				}
			}
	      		free (data.dptr);
		}	
		gdbm_close(pw_file);
	}
}

void DeleteAccount()
{
	DoHeader("Password Administration");
	fprintf(cgiOut,"<H1 align=center>Password Administration - Delete an account</H1>\n");

	char name[MaxUserNameLength];
	char AdminPass[MaxUserNameLength];
	*name=*AdminPass='\0';

	datum key, data;	
	
	cgiFormStringNoNewlines("name", name, MaxUserNameLength);
	cgiFormStringNoNewlines("pass", AdminPass, MaxUserNameLength);

	if (!*name || !*AdminPass)
	{
		fprintf(cgiOut,	"You did not fill in all the fields."
				" Please use the back button on your web browser to return to the form.");
	}
	else
	{
		GDBM_FILE pw_file = gdbm_open(PasswordFileName, 0, GDBM_WRCREAT, 00664, NULL);					
		if (!pw_file)
		{
			fprintf(cgiOut,	"Error: could not open database.\n");
			return;
		}

		datum key, data;	
		UserInfo user;

		// verify that admin's password was entered correctly
		key.dptr="admin";
		key.dsize=strlen(key.dptr) + 1;
		data = gdbm_fetch(pw_file, key);
		if (data.dptr != NULL)
		{
			UserInfo *admin=NULL;
			admin = (UserInfo *) data.dptr;		
			char adminPassDB[MaxUserNameLength];
			strcpy(adminPassDB, admin->pass);
	      		free (data.dptr);
			if (strcmp(adminPassDB, wrapCrypt(AdminPass)))
			{
				fprintf(cgiOut, "Wrong password.\n");
			}
			else
			{
				key.dptr=name;
				key.dsize=strlen(key.dptr) + 1;				
				if (!gdbm_delete (pw_file, key))
				{
					fprintf(cgiOut, "Successfully deleted %s's account", name);
				}
				else
				{
					fprintf(cgiOut, "Could not delete %s's account.  Perhaps it doesn't exist.", name);
				}
			}
		}
		gdbm_close(pw_file);
	}
}

void DeleteOld()
{
	DoHeader("Password Administration");
	fprintf(cgiOut,"<H1 align=center>Password Administration - Delete old accounts</H1>\n");

	char AdminPass[MaxUserNameLength];
	*AdminPass='\0';
	
	cgiFormStringNoNewlines("pass", AdminPass, MaxUserNameLength);

	if (!*AdminPass)
	{
		fprintf(cgiOut,	"You did not fill in all the fields."
				" Please use the back button on your web browser to return to the form.");
	}
	else
	{
		GDBM_FILE pw_file = gdbm_open(PasswordFileName, 0, GDBM_WRCREAT, 00664, NULL);					
		if (!pw_file)
		{
			fprintf(cgiOut,	"Error: could not open database.\n");
			return;
		}

		datum key, data;	
		UserInfo user;

		// verify that admin's password was entered correctly
		key.dptr="admin";
		key.dsize=strlen(key.dptr) + 1;
		data = gdbm_fetch(pw_file, key);
		if (data.dptr != NULL)
		{
			UserInfo *admin=NULL;
			admin = (UserInfo *) data.dptr;		
			char adminPassDB[MaxUserNameLength];
			strcpy(adminPassDB, admin->pass);
	      		free (data.dptr);
			if (strcmp(adminPassDB, wrapCrypt(AdminPass)))
			{
				fprintf(cgiOut, "Wrong password.\n");
			}
			else
			{
				unsigned int RecordsDeleted=0;
				// 60 days * 24 hours * 60 minutes * 60 seconds
				time_t OldestTime = time(NULL) - (60 * 24 * 60 * 60);
				// loop through the entire database, deleting old accounts
				key = gdbm_firstkey ( pw_file );
				while (key.dptr)
				{
					data = gdbm_fetch(pw_file, key);					
					UserInfo *tempUser = (UserInfo *) data.dptr;
					if (tempUser->lastLogin < OldestTime)
					{
						// delete this record
						gdbm_delete (pw_file, key);
						RecordsDeleted++;
					}
					
					free (data.dptr);
					free (key.dptr);
					key = gdbm_nextkey( pw_file, key );
				}
				fprintf(cgiOut, "Deleted %d old accounts", RecordsDeleted);
			}
		}
		gdbm_close(pw_file);
	}
}

// ReadMessageHeaders()
// Call with pointer to opened binary file (MessageFile) and
// address of pointer to MessageHeader (messages).
// Memory for (messages) will be allocated with new.
// The array (messages) will be sorted by message number.
// Number of messages read will be returned.
unsigned ReadMessageHeaders(FILE *MessageFile, MessageHeader **messages)
{
	*messages=NULL;
	fseek(MessageFile, 0L, SEEK_END);
	unsigned count = ftell(MessageFile) / sizeof(MessageHeader);
	if (count > 0)
	{
		*messages = new MessageHeader[count];
		fseek(MessageFile, 0L, SEEK_SET);
		fread(*messages, sizeof(MessageHeader), count, MessageFile);
       		qsort(*messages, count, sizeof(MessageHeader),compareMessageHeader);
	}
	return count;
}

const char *displayTime(time_t timeIn)
{
	struct tm *st = localtime(&timeIn);
	static char stringTime[50];
	char ap='a';
	int hour=st->tm_hour;
	if (hour > 12)
	{
		ap='p';
		hour-=12;
	}

	sprintf(stringTime, "%d/%d/%d %d:%02d %cm",
		st->tm_mon+1, st->tm_mday, st->tm_year, hour, st->tm_min, ap);
	return stringTime;
}

// Show the messages as threads.
void showMessages(const char *room, const MessageHeader *messages, unsigned count, unsigned thread=0)
{
	unsigned currentMessage=0;
	int inList=0;
	fprintf(cgiOut, "<UL>\n");
	for (unsigned index=0; index < count; index++)
	{	const MessageHeader &m = messages[index];	
		if (m.number.message) // skip messages that were deleted by search()
		{
			if (!thread || m.number.message == thread )
			{
				if (m.number.message == currentMessage)
				{
					if (!inList)
					{
						fprintf(cgiOut, "<UL>\n");
						inList=1;
					}
				}
				else if (inList)
				{
					fprintf(cgiOut, "</UL>\n");
					inList=0;
				}
				fprintf(cgiOut, "<LI><A HREF=\"%s?action=read&forum=%s&message=%d-%d\">%s</A> %s (%s)",
						cgiURL, room, m.number.message,
						m.number.reply,	strlen(m.subject) ? m.subject : "No Subject", m.userName,
						displayTime(m.timePosted)); 
				currentMessage = m.number.message;
			}
		}
	}
	if (inList)
	{
		fprintf(cgiOut, "</UL>\n");
	}
	fprintf(cgiOut, "</UL>\n");
}

void showPostForm(const char *forum)
{
	DoHeader("Post a message");
	char HeaderFileName[MaxFileNameLength];			
	sprintf(HeaderFileName, "%s.ph", forum);
	ShowFile(commonHeaderFileName);
	ShowFile(HeaderFileName);	
	int nReplyNum;
	cgiFormInteger("number", &nReplyNum, 0);		
	const char *temp = nReplyNum ? "reply" : "new message";
	fprintf(cgiOut, "<H3>Posting a %s</H3>", temp);
	fprintf(cgiOut, "<FORM action=\"%s\" method=post>\n",cgiURL);
	fprintf(cgiOut, "<input type=hidden name=forum value=\"%s\">\n", forum);
	fprintf(cgiOut, "<input type=hidden name=number value=\"%d\">\n", nReplyNum);
	fprintf(cgiOut, "<TABLE BORDER=0>\n");
	fprintf(cgiOut, "<TR><TH align=left>User Name:</TH><TD><input name=user maxlength=%d size=%d><BR></TD></TR>\n", MaxUserNameLength, MaxUserNameLength);
	fprintf(cgiOut, "<TR><TH align=left>Password:</TH><TD><input type=password name=pass maxlength=%d size=%d><BR></TD></TR>\n", MaxUserNameLength, MaxUserNameLength);
	char s[MaxSubjectLength], s2[MaxSubjectLength + 4];
	*s2=*s='\0';
	if (nReplyNum)
	{
		cgiFormStringNoNewlines("subject", s,MaxSubjectLength);
		sprintf(s2, "RE: %s", s);
	}
	fprintf(cgiOut, "<TR><TH align=left>Subject:</TH><TD><input name=subject maxlength=%d size=%d value=\"%s\"><BR></TD></TR>\n",MaxSubjectLength, 60, s2);
	fprintf(cgiOut, "<TR><TH align=left>Body:</TH><TD><textarea name=body rows=10 cols=60 WRAP=VIRTUAL></textarea><BR></TD></TR>\n");
	fprintf(cgiOut, "</TABLE>\n");
	fprintf(cgiOut, "Check here if your message is pre-formatted (i.e. poetry): <input type=CHECKBOX name=pre><BR>");
	fprintf(cgiOut, "<input name=action type=submit value=\"post\"> or ");
	fprintf(cgiOut, "<input name=action type=submit value=\"preview\"> this message<BR>\n");
	fprintf(cgiOut, "</FORM>");
}

void DisplayMessage(const char *pRoom, const char *pUserName, time_t nTime, const char *pSubject, const char *pBody)
{
	char HeaderFileName[MaxFileNameLength];			
	sprintf(HeaderFileName, "%s.mh", pRoom);
	ShowFile(commonHeaderFileName);
	ShowFile(HeaderFileName);		
	fprintf(cgiOut, "<TABLE BORDER=0>\n");
	fprintf(cgiOut, "<TR><TH align=left>Name:</TH><TD>%s<BR></TD></TR>\n", pUserName);
	fprintf(cgiOut, "<TR><TH align=left>Date/Time:</TH><TD>%s<BR></TD></TR>\n", displayTime(nTime));
	fprintf(cgiOut, "<TR><TH align=left>Subject:</TH><TD>%s<BR></TD></TR>\n", 
			strlen(pSubject) ? pSubject : "No Subject");
	fprintf(cgiOut, "</TABLE>\n");
	fprintf(cgiOut, "<BR><B>Body:</B><BLOCKQUOTE>\n");
	fprintf(cgiOut, "%s", pBody);
	fprintf(cgiOut, "</BLOCKQUOTE>\n");
}

void post(const char *pRoom)
{
	char pUserName[MaxUserNameLength];
	char pUserPass[MaxUserNameLength];
	char pSubject[MaxSubjectLength];
	int nReplyNum=0;
	*pUserName=*pUserPass=*pSubject;
	cgiFormStringNoNewlines("user", pUserName, MaxUserNameLength);
	cgiFormStringNoNewlines("pass", pUserPass, MaxUserNameLength);
	if (VerifyPassword(pUserName, pUserPass, 0))
	{
		cgiFormInteger("number", &nReplyNum, 0);
		cgiFormStringNoNewlines("subject", pSubject, MaxSubjectLength);
//		String strSubject = stripHTML(pSubject, MaxSubjectLength);
		stripHtmlInPlace(pSubject);
		int pre = (cgiFormSuccess == cgiFormCheckboxSingle("pre"));
		int nBodyLength=0;
		cgiFormStringSpaceNeeded("body", &nBodyLength);
		String strMessageBody="";
		char *pMessageBody=NULL;
		if (nBodyLength)
		{
			pMessageBody = new char[nBodyLength + 1];
			*pMessageBody=0;
			cgiFormString("body", pMessageBody, nBodyLength + 1);	
			strMessageBody = pMessageBody;
		}
		
		if (pre) 
		{
			// if they want it formatted exactly like they typed it
			cat("<PRE>\n", strMessageBody, "\n</PRE>\n", strMessageBody);
		}
		else
		{	
			// Make it look nice
			strMessageBody.gsub("\n", "<BR>\n");
		}

		DoHeader("Message Preview");
		DisplayMessage(pRoom, pUserName, time(NULL), pSubject, strMessageBody);
		fprintf(cgiOut, "<HR>\n");
		fprintf(cgiOut, "Use the back button on your browser to make corrections and/or post your message.<BR>\n");
		if (pMessageBody)
		{
			delete[] pMessageBody;
			pMessageBody=NULL;
		}
	}
	else
	{
		WrongPassword();
	}
}

void confirmPost(const char *forum)
{
	char pUserName[MaxUserNameLength];
	char pUserPass[MaxUserNameLength];
	char pSubject[MaxSubjectLength];
	*pUserName=*pUserPass=*pSubject;
	int nReplyNum=0;
	cgiFormStringNoNewlines("user", pUserName, MaxUserNameLength);
	cgiFormStringNoNewlines("pass", pUserPass, MaxUserNameLength);
	if (VerifyPassword(pUserName, pUserPass, 0))
	{
		String strMessageBody;
		cgiFormInteger("number", &nReplyNum, 0);
		cgiFormStringNoNewlines("subject", pSubject, MaxSubjectLength);
//		String strSubject = stripHTML(pSubject, MaxSubjectLength);
		stripHtmlInPlace(pSubject);
		
		int nBodyLength=0;
		cgiFormStringSpaceNeeded("body", &nBodyLength);
		char *pMessageBody = NULL;
		if (nBodyLength)
		{
			pMessageBody = new char[nBodyLength + 1];
			*pMessageBody=0;
			cgiFormString("body", pMessageBody, nBodyLength + 1);	
			strMessageBody = pMessageBody;
			delete[] pMessageBody;
			pMessageBody=NULL;
			int pre = (cgiFormSuccess == cgiFormCheckboxSingle("pre"));
			if (pre) 
			{
				// if they want it formatted exactly like they typed it
				cat("<PRE>\n", strMessageBody, "\n</PRE>\n", strMessageBody);
			}
			else
			{	
				// Make it look nice
				strMessageBody.gsub("\n", "<BR>\n");
			}
		}

		char MessageFileName[MaxFileNameLength];
		sprintf(MessageFileName, "%s.msg", forum);
		GDBM_FILE gdbmFile = gdbm_open(MessageFileName, 0, GDBM_WRCREAT, 00664, NULL);					
		if (!gdbmFile)
		{
			String strErrorMessage;
			cat("Error: could not open ", forum, " message database.", strErrorMessage);
			DoError(strErrorMessage);
		}
		else
		{
			String strLockName=forum;
			strLockName += ".loc";
			Lock MyLock(strLockName, 10);
			if (MyLock.locked())
			{
				MessageHeader *messages=NULL;
				char MessageFileName[MaxFileNameLength];
				sprintf(MessageFileName, "%s.hdr", forum);
				FILE *MessageFile = fopen(MessageFileName, "r+b");
				unsigned count=0;
				if (MessageFile)
				{
					count = ReadMessageHeaders(MessageFile, &messages);
				}
				else
				{
					MessageFile = fopen(MessageFileName, "wb");
				}
				
				
				if (MessageFile)
				{
					MessageNumber mn;	
					if (nReplyNum)
					{
						mn.message = nReplyNum;
						mn.reply=0;
						for (long pos=count-1; pos > -1; pos--)
						{
							if (messages[pos].number.message == mn.message)
							{
								mn.reply = messages[pos].number.reply + 1;
								break;
							}
						}
						if (!mn.reply)
						{
							// didn't find it
							mn.message = messages[count-1].number.message + 1;
							mn.reply=1;
						}					
					}
					else
					{
						if (messages)
						{
							mn.message = messages[count-1].number.message + 1;
						}
						else
						{
							mn.message=1;
						}
						mn.reply=1;
					}
								
					datum key, data;
					key.dptr=(char *) &mn;
					key.dsize=sizeof(MessageNumber);
		
					MessageHeader mh;
					strcpy(mh.userName, pUserName);
					mh.timePosted=time(NULL);
					strcpy(mh.subject, pSubject);
					mh.number.message=mn.message;
					mh.number.reply=mn.reply;
					
					fseek(MessageFile, 0, SEEK_END);
					fwrite(&mh, sizeof(MessageHeader), 1, MessageFile);
					fclose(MessageFile);
					if (!strMessageBody.length())
					{
						strMessageBody = "No message";
					}				
					unsigned tempSize = sizeof(MessageHeader) + strMessageBody.length() + 1;
					char *temp = (char *) malloc(tempSize);
					memcpy(temp, &mh, sizeof(MessageHeader));
					strcpy(&(temp[sizeof(MessageHeader)]), (const char *)strMessageBody);
					data.dptr = temp;
					data.dsize = tempSize;
					gdbm_store(gdbmFile, key, data, GDBM_REPLACE);
					free(temp);
					temp=NULL;
					gdbm_close(gdbmFile);
				}
				else
				{
					DoError("Error: Could not open the message database");
				}
			}
			else
			{
				DoError("Error: Could not lock message database");
			}			
		}
	}
	else
	{
		WrongPassword();
	}
	
	DoHeader("Message Posted");
	fprintf(cgiOut, "Your message was posted.<BR>");
	fprintf(cgiOut, "<FORM action=\"%s\" method=post>\n",cgiURL);
	fprintf(cgiOut, "<input type=hidden name=forum value=\"%s\">\n", forum);
	fprintf(cgiOut, "<input type=submit value=\"Return to the index\">\n");
	fprintf(cgiOut, "</FORM>");

}

void read(const char *forum)
{
	char MessageFileName[MaxFileNameLength];
	sprintf(MessageFileName, "%s.msg", forum);
	GDBM_FILE gdbmFile = gdbm_open(MessageFileName, 0, GDBM_READER, 00664, NULL);					
	if (!gdbmFile)
	{
		const char *fmt="Error: could not open %s message database.";
		char *ErrorMessage = new char[strlen(fmt) + MaxFileNameLength + 1];
		sprintf(ErrorMessage, fmt, forum);
		DoError(ErrorMessage);
		delete[] ErrorMessage;
	}
	else
	{
		int messageNumberLength=0;
		cgiFormStringSpaceNeeded("message", &messageNumberLength);
		if (!messageNumberLength)
		{
			DoError("Error: No message number supplied");
			return;
		}
		char *strMessageNumber = new char[messageNumberLength + 1];
		*strMessageNumber=0;
		cgiFormStringNoNewlines("message", strMessageNumber, messageNumberLength + 1);
		MessageNumber mn;
		sscanf(strMessageNumber, "%d-%d", &(mn.message), &(mn.reply));
		delete[] strMessageNumber;
		strMessageNumber=NULL;
		datum key, data;
		key.dptr=(char *)&mn;
		key.dsize=sizeof(MessageNumber);
		data = gdbm_fetch(gdbmFile, key);
		if (data.dptr)
		{
			MessageHeader *mh = (MessageHeader*) data.dptr;
			char *body = data.dptr + sizeof(MessageHeader);
			DoHeader("Reading a message");
			DisplayMessage(forum, mh->userName, mh->timePosted, mh->subject, body);
			fprintf(cgiOut, "<H3>Messages in this thread:</H3>");
			// Show messages in this thread
			char MessageHeaderFileName[MaxFileNameLength];
			sprintf(MessageHeaderFileName, "%s.hdr", forum);
			FILE *MessageFile = fopen(MessageHeaderFileName, "rb");
			if (MessageFile)
			{	
				MessageHeader *messages=NULL;
				unsigned count = ReadMessageHeaders(MessageFile, &messages);
			    	showMessages(forum, messages, count, mh->number.message);
				delete[] messages;
				messages=NULL;
			}	
			
			stripHtmlInPlace(mh->subject);
			fprintf(cgiOut, "<HR><CENTER> <FORM METHOD=POST ACTION=\"%s\">"
					"<input type=hidden name=\"number\" value=\"%d\">"
					"<input type=hidden name=\"forum\" value=\"%s\">"
					"<input type=hidden name=\"action\" value=\"postform\">"
					"<input type=hidden name=\"subject\" value=\"%s\">"
					"<input type=submit value=\"Post a reply\">"
					"</FORM>\n",
					cgiURL, mh->number.message, forum, mh->subject);

			fprintf(cgiOut, "<FORM METHOD=POST ACTION=\"%s\">"
					"<input type=hidden name=\"forum\" value=\"%s\">"
					"<input type=hidden name=\"action\" value=\"postform\">"
					"<input type=submit value=\"Post a new message\">"
					"</FORM>\n",
					cgiURL, forum);

			fprintf(cgiOut, "If you posted this message, you can:<BR>");
			fprintf(cgiOut, "<FORM METHOD=POST ACTION=\"%s\">"
					"<input type=hidden name=\"message\" value=\"%d-%d\">"
					"<input type=hidden name=\"forum\" value=\"%s\">"
					"<input type=hidden name=\"action\" value=\"delete\">"
					"Password: <input type=password name=\"pass\"> "
					"<input type=submit value=\"Delete this message\">"
					"</FORM></CENTER>\n",
					cgiURL, mh->number.message, mh->number.reply, 
					forum, mh->subject);
			free(data.dptr);		
		}
		else
		{
			fprintf(cgiOut, "That message doesn't seem to be here.  Perhaps it's been deleted.<P>");
		}
		gdbm_close(gdbmFile);
	}
}

WriteMessageHeaders(FILE *MessageFile, MessageHeader *messages, unsigned count)
{
	fseek(MessageFile, 0L, SEEK_SET);
	for (unsigned int x=0; x<count; x++)
	{
		if (messages[x].number.message)
		{
			fwrite(&(messages[x]), sizeof(MessageHeader), 1, MessageFile);		
		}
	}
}

void deleteMessage(MessageHeader *target, GDBM_FILE gdbmFile)
{
	// delete the message
	datum key;
	key.dptr=(char *)&(target->number);
	key.dsize=sizeof(MessageNumber);
	gdbm_delete(gdbmFile, key);
	
	// mark the header as deleted so we won't write it back out
	target->number.message=0;
}

void deleteOldMessages(const char *forum, unsigned daysOld)
{
	if (!daysOld || !forum)
	{
		return;
	}
	char HeaderFileName[MaxFileNameLength];
	sprintf(HeaderFileName, "%s.hdr", forum);
	char MessageFileName[MaxFileNameLength];
	sprintf(MessageFileName, "%s.msg", forum);
	GDBM_FILE gdbmFile = gdbm_open(MessageFileName, 0, GDBM_WRITER, 00664, NULL);					
	if (gdbmFile)
	{
		String strLockName=forum;
		strLockName += ".loc";
		Lock MyLock(strLockName, 10);
		if (MyLock.locked())
		{
			FILE *MessageFile = fopen(HeaderFileName, "rb");
			if (MessageFile)
			{
				MessageHeader *messages;
				unsigned count = ReadMessageHeaders(MessageFile, &messages);
				fclose(MessageFile);
				time_t oldestTime = time(NULL) - (daysOld * 24 * 60 * 60);
				int deletedSomething=0;
				for (unsigned x=0; x<count; x++)
				{
					if (messages[x].timePosted < oldestTime)
					{
						deleteMessage(&(messages[x]), gdbmFile);
						deletedSomething=1;
					}
				}
				if (deletedSomething)
				{
					FILE *MessageFile = fopen(HeaderFileName, "wb");
					if (MessageFile)
					{
						WriteMessageHeaders(MessageFile, messages, count);
						fclose(MessageFile);
					}
				}
				delete[] messages;
			}
		}
		gdbm_close(gdbmFile);
	}
}

void deleteMessage(const char *forum)
{
	char HeaderFileName[MaxFileNameLength];
	sprintf(HeaderFileName, "%s.hdr", forum);
	char MessageFileName[MaxFileNameLength];
	sprintf(MessageFileName, "%s.msg", forum);
	char ForumHeaderName[MaxFileNameLength];
	sprintf(ForumHeaderName, "%s.mh", forum);
	GDBM_FILE gdbmFile = gdbm_open(MessageFileName, 0, GDBM_WRITER, 00664, NULL);					

	if (!gdbmFile)
	{
		const char *fmt="Error: could not open %s message database.";
		char *ErrorMessage = new char[strlen(fmt) + MaxFileNameLength + 1];
		sprintf(ErrorMessage, fmt, forum);
		DoError(ErrorMessage);
		delete[] ErrorMessage;
	}
	else
	{
		int messageNumberLength=0;
		cgiFormStringSpaceNeeded("message", &messageNumberLength);
		if (!messageNumberLength)
		{
			DoError("Error: No message number supplied");
			return;
		}
		char *strMessageNumber = new char[messageNumberLength + 1];
		*strMessageNumber=0;
		cgiFormStringNoNewlines("message", strMessageNumber, messageNumberLength + 1);
		MessageHeader target;
		sscanf(strMessageNumber, "%d-%d", &(target.number.message), &(target.number.reply));
		delete[] strMessageNumber;
		strMessageNumber=NULL;

		String strLockName=forum;
		strLockName += ".loc";
		Lock MyLock(strLockName, 10);
		if (MyLock.locked())
		{
			FILE *MessageFile = fopen(HeaderFileName, "rb");
			if (MessageFile)
			{
				MessageHeader *messages;
				unsigned count = ReadMessageHeaders(MessageFile, &messages);
				fclose(MessageFile);
				MessageHeader *found=(MessageHeader *) bsearch(&target, messages, count, sizeof(MessageHeader), compareMessageHeader);
				if (found)
				{
					char pUserPass[MaxUserNameLength];
					cgiFormStringNoNewlines("pass", pUserPass, MaxUserNameLength);
					if (	!VerifyPassword(found->userName, pUserPass, 1) &&
						!VerifyPassword("admin", pUserPass, 1))
					{
						WrongPassword();
					}
					else
					{
						deleteMessage(found, gdbmFile);
						FILE *MessageFile = fopen(HeaderFileName, "wb");
						if (MessageFile)
						{
							WriteMessageHeaders(MessageFile, messages, count);
							fclose(MessageFile);
						}
						DoHeader("Message deleted");
						ShowFile(commonHeaderFileName);
						ShowFile(ForumHeaderName);		
						fprintf(cgiOut, "The message has been deleted.");
					}
				}
				else
				{
					DoHeader("Not found");
					ShowFile(commonHeaderFileName);
					ShowFile(ForumHeaderName);		
					fprintf(cgiOut, "Sorry, that message doesn't seem to be here.  "
							"Maybe it's already been deleted.");
				}
				delete[] messages;
			}
		}
		gdbm_close(gdbmFile);
	}
}

// Filter the list of messages based on search criterea
// (max days old, username contains, subject contains), 
// if supplied.
void filterMessages(MessageHeader *messages, unsigned count)
{
	int daysOld;
	cgiFormInteger("searchdays", &daysOld, 0);
	char user[MaxUserNameLength];
	*user=0;
	cgiFormStringNoNewlines("searchuser", user, MaxUserNameLength);
	char subject[MaxSubjectLength];
	*subject=0;
	cgiFormStringNoNewlines("searchsubject", subject, MaxSubjectLength);
	
	if (daysOld || *user || *subject) // Filter the listing if they provided some criterea
	{
	       	time_t oldestTime = time(NULL) - (daysOld * 24 * 60 * 60);
		WBBS_strupr(user); // convert to upercase
		WBBS_strupr(subject);
		// set message number = 0 for messages that don't match the search criteria	        
		for (int i=0; i<count; i++)
		{
			if (daysOld && messages[i].timePosted < oldestTime)
			{
				messages[i].number.message = 0;
			}
			
			if (*user)
			{
				char messageName[MaxUserNameLength];
				strcpy(messageName, messages[i].userName);
				WBBS_strupr(messageName);		
				if (!strstr(messageName, user))
				{	
					messages[i].number.message = 0;
				}
			}
			
			if (*subject)
			{
				char messageSubject[MaxSubjectLength];
				strcpy(messageSubject, messages[i].subject);
				WBBS_strupr(messageSubject);		
				if (!strstr(messageSubject, subject))
				{	
					messages[i].number.message = 0;
				}
			}
		}
	}
}

// Show the index of messages
void index(const char *room)
{
	char HeaderFileName[MaxFileNameLength];
	char MessageFileName[MaxFileNameLength];
	sprintf(HeaderFileName, "%s.ih", room);
	sprintf(MessageFileName, "%s.hdr", room);
	
	DoHeader("Index of messages");
	ShowFile(commonHeaderFileName);
	ShowFile(HeaderFileName);
	MessageHeader *messages=NULL;
	FILE *MessageFile = fopen(MessageFileName, "rb");
	if (MessageFile)
	{
		unsigned count = ReadMessageHeaders(MessageFile, &messages);
		filterMessages(messages, count);
	       	showMessages(room, messages, count);
		delete[] messages;
	}
	fprintf(cgiOut, "<HR><CENTER> <FORM METHOD=POST ACTION=\"%s\">"
			"<input type=hidden name=\"action\" value=\"postform\">"
			"<input type=hidden name=\"forum\" value=\"%s\">"
			"<input type=submit value=\"Post a message\">"
			"</FORM></CENTER>",
			cgiURL, room);
}

void finishedPasswordAdmin()
{
	fprintf(cgiOut, "<P><A HREF=\"%s\">Return to the forums</A>\n", EntraceURL);
}

int GetForumName(char forum[MaxFileNameLength])
{
	int retVal=1;
        cgiFormStringNoNewlines("forum", forum, MaxFileNameLength);
	if (VerifyRoomName(forum))
	{
		deleteOldMessages(forum, 30);
	}
	else
	{
		DoError("Invalid forum name");
		retVal=0;
	}
	return retVal;
}

int cgiMain() 
{
        char action[MaxFileNameLength];
	char forum[MaxFileNameLength];
	*action=*forum='\0';
        cgiFormStringNoNewlines("action", action, MaxFileNameLength);
		
	if (!strcmp(action, "read"))
	{
		// read a message
		if (GetForumName(forum))
		{
			read(forum);
		}
	}
	else if (!strcmp(action, "delete"))
	{
		// delete a message
		if (GetForumName(forum))
		{
			deleteMessage(forum);
		}
	}
	else if (!strcmp(action, "postform"))
	{
		// Show the form for posting in this forum
		if (GetForumName(forum))
		{
			showPostForm(forum);
		}
	}
	else if (!strcmp(action, "preview"))
	{
		// post a message, just echos it back for confirmation
		if (GetForumName(forum))
		{
			post(forum);
		}
	}
	else if (!strcmp(action, "post"))
	{
		// post a message, for real
		if (GetForumName(forum))
		{
			confirmPost(forum);
		}
	}
	else if (!strcmp(action, "create_account"))
	{
		CreateAccount();				
		finishedPasswordAdmin();
	} else if (!strcmp(action, "change_password"))
	{
		ChangePassword();	
		finishedPasswordAdmin();
	} else if (!strcmp(action, "delete_account"))
	{
		DeleteAccount();	
		finishedPasswordAdmin();
	} else if (!strcmp(action, "delete_old"))
	{
		DeleteOld();	
		finishedPasswordAdmin();
	}
	else
	{
		// Default action is to show index of messages.
		if (GetForumName(forum))
		{
			index(forum);
		}
	}
	fprintf(cgiOut,	"</BODY></HTML>\n");
	return 0;
}




